import argparse
import sys
import os

from setuptools import find_packages, setup

# pylint: disable=E0401, W0611
if sys.version_info[0] < 3:
    import __builtin__ as builtins
else:
    import builtins


def long_description():
    here = os.path.abspath(os.path.dirname(__file__))
    with open(os.path.join(here, 'README.rst'), 'r') as fh:
        return fh.read()


version = {}
with open("dagster/version.py") as fp:
    exec(fp.read(), version)  # pylint: disable=W0122


parser = argparse.ArgumentParser()
parser.add_argument('--nightly', action='store_true')


def _do_setup(name='dagster'):
    setup(
        name=name,
        version=version['__version__'],
        author='Elementl',
        license='Apache-2.0',
        description='Dagster is an opinionated programming model for data pipelines.',
        long_description=long_description(),
        long_description_content_type='text/markdown',
        url='https://github.com/dagster-io/dagster',
        classifiers=[
            'Programming Language :: Python :: 2.7',
            'Programming Language :: Python :: 3.5',
            'Programming Language :: Python :: 3.6',
            'License :: OSI Approved :: Apache Software License',
            'Operating System :: OS Independent',
        ],
        packages=find_packages(exclude=['dagster_tests']),
        install_requires=[
            # standard python 2/3 compatability things
            'enum34>=1.1.6',
            'future>=0.16.0',
            'funcsigs>=1.0.2',
            'contextlib2>=0.5.5',
            # cli
            'click>=6.7',
            'coloredlogs>=10.0',
            'graphviz>=0.8.3',
            'pyyaml>=3.12',
            # core (not explicitly expressed atm)
            'six>=1.11.0',
            'toposort>=1.5',
            # dev/test - Installed via dev-requirements.txt
            # 'pylint>=1.8.4',
            # 'pytest>=3.5.1',
            # 'recommonmark>=0.4.0',
            # 'rope>=0.10.7',
            # 'Sphinx>=1.7.5',
            # 'sphinx-autobuild>=0.7.1',
            # 'yapf>=0.22.0',
            # 'twine>=1.11.0',
            # 'pre-commit'>=1.10.1',
        ],
        extras_require={":python_version>'3'": ["reloader>=0.6"]},
        entry_points={"console_scripts": ['dagster = dagster.cli:main']},
    )


if __name__ == '__main__':
    parsed, unparsed = parser.parse_known_args()
    sys.argv = [sys.argv[0]] + unparsed
    if parsed.nightly:
        _do_setup('dagster-nightly')
    else:
        _do_setup('dagster')



from dagster import check

from .config import resolve_to_config_type


class InputSchema:
    @property
    def schema_cls(self):
        check.not_implemented('Must implement')

    @property
    def schema_type(self):
        return resolve_to_config_type(self.schema_cls)

    def construct_from_config_value(self, value):
        return value


def make_input_schema(dagster_type):
    class _InputSchema(InputSchema):
        @property
        def schema_cls(self):
            return dagster_type

    return _InputSchema()


class OutputSchema:
    @property
    def schema_cls(self):
        check.not_implemented('Must implement')

    @property
    def schema_type(self):
        return resolve_to_config_type(self.schema_cls)

    def materialize_runtime_value(self, _config_value, _runtime_value):
        check.not_implemented('Must implement')



import json

from dagster import check

from .builtin_enum import BuiltinEnum
from .config import ConfigTypeAttributes
from .config_schema import OutputSchema
from .field import Field, Dict, NamedSelector


def define_path_dict_field():
    return Field(Dict({'path': Field(BuiltinEnum.PATH)}))


def define_builtin_scalar_output_schema(scalar_name):
    check.str_param(scalar_name, 'scalar_name')

    schema_cls = NamedSelector(
        scalar_name + '.MaterializationSchema',
        {'json': define_path_dict_field()},
        type_attributes=ConfigTypeAttributes(is_system_config=True),
    )

    class _BuiltinScalarOutputSchema(OutputSchema):
        @property
        def schema_cls(self):
            return schema_cls

        def materialize_runtime_value(self, config_spec, runtime_value):
            check.dict_param(config_spec, 'config_spec')
            selector_key, selector_value = list(config_spec.items())[0]

            if selector_key == 'json':
                json_file_path = selector_value['path']
                json_value = json.dumps({'value': runtime_value})
                with open(json_file_path, 'w') as ff:
                    ff.write(json_value)
            else:
                check.failed(
                    'Unsupported selector key: {selector_key}'.format(selector_key=selector_key)
                )

    return _BuiltinScalarOutputSchema()



from toposort import toposort_flatten

from dagster import check

from .context import PipelineContextDefinition, default_pipeline_context_definitions

from .dependency import DependencyDefinition, DependencyStructure, Solid

from .pipeline_creation import (
    create_execution_structure,
    construct_config_type_dictionary,
    construct_runtime_type_dictionary,
)

from .utils import check_opt_two_dim_dict


class PipelineDefinition(object):
    '''A instance of a PipelineDefinition represents a pipeline in dagster.

    A pipeline is comprised of:

    - Solids:
        Each solid represents a functional unit of data computation.

    - Context Definitions:
        Pipelines can be designed to execute in a number of different operating environments
        (e.g. prod, dev, unittest) that require different configuration and setup. A context
        definition defines how a context (of type ExecutionContext) is created and what
        configuration is necessary to create it.

    - Dependencies:
        Solids within a pipeline are arranged as a DAG (directed, acyclic graph). Dependencies
        determine how the values produced by solids flow through the DAG.

    Attributes:
        name (str):
            Name of the pipeline. Must be unique per-repository.
        description (str):
            Description of the pipeline. Optional.
        solids (List[SolidDefinition]):
            List of the solids in this pipeline.
        dependencies (Dict[str, Dict[str, DependencyDefinition]]) :
            Dependencies that constitute the structure of the pipeline. This is a two-dimensional
            array that maps solid_name => input_name => DependencyDefiniion instance
        context_definitions (Dict[str, PipelineContextDefinition]):
            The context definitions available for consumers of this pipelines. For example, a
            unit-testing environment and a production environment probably have very different
            configuration and requirements. There would be one context definition per
            environment.
        dependency_structure (DependencyStructure):
            Used mostly internally. This has the same information as the dependencies data
            structure, but indexed for fast usage.
    '''

    def __init__(
        self, solids, name=None, description=None, context_definitions=None, dependencies=None
    ):
        '''
        Args:
            solids (List[SolidDefinition]): Solids in the pipeline
            name (str): Name. This is optional, mostly for situations that require ephemeral
                pipeline definitions for fast scaffolding or testing.
            description (str): Description of the pipline.
            context_definitions (Dict[str, PipelineContextDefinition]): See class description.
            dependencies: (Dict[str, Dict[str, DependencyDefinition]]): See class description.
        '''
        self.name = check.opt_str_param(name, 'name', '<<unnamed>>')
        self.description = check.opt_str_param(description, 'description')

        check.list_param(solids, 'solids')

        if context_definitions is None:
            context_definitions = default_pipeline_context_definitions()

        self.context_definitions = check.dict_param(
            context_definitions,
            'context_definitions',
            key_type=str,
            value_type=PipelineContextDefinition,
        )

        self.dependencies = check_opt_two_dim_dict(
            dependencies, 'dependencies', value_type=DependencyDefinition
        )

        dependency_structure, pipeline_solid_dict = create_execution_structure(
            solids, self.dependencies
        )

        self._solid_dict = pipeline_solid_dict
        self.dependency_structure = dependency_structure

        from dagster.core.system_config.types import define_environment_cls

        self.environment_cls = define_environment_cls(self)
        self.environment_type = self.environment_cls.inst()

        self._config_type_dict = construct_config_type_dictionary(
            solids, self.context_definitions, self.environment_type
        )

        self._runtime_type_dict = construct_runtime_type_dictionary(solids)

    @property
    def display_name(self):
        '''Name suitable for exception messages, logging etc. If pipeline
        is unnamed the method with return "<<unnamed>>".

        Returns:
            str: Display name of pipeline
        '''
        return self.name if self.name else '<<unnamed>>'

    @property
    def solids(self):
        '''Return the solids in the pipeline.

        Returns:
            List[SolidDefinition]: List of solids.
        '''
        return list(set(self._solid_dict.values()))

    def has_solid(self, name):
        '''Return whether or not the solid is in the piepline

        Args:
            name (str): Name of solid

        Returns:
            bool: True if the solid is in the pipeline
        '''
        check.str_param(name, 'name')
        return name in self._solid_dict

    def solid_named(self, name):
        '''Return the solid named "name". Throws if it does not exist.

        Args:
            name (str): Name of solid

        Returns:
            SolidDefinition: SolidDefinition with correct name.
        '''
        check.str_param(name, 'name')
        return self._solid_dict[name]

    def has_config_type(self, name):
        check.str_param(name, 'name')
        return name in self._config_type_dict

    def config_type_named(self, name):
        check.str_param(name, 'name')
        return self._config_type_dict[name]

    def has_runtime_type(self, name):
        check.str_param(name, 'name')
        return name in self._runtime_type_dict

    def runtime_type_named(self, name):
        check.str_param(name, 'name')
        return self._runtime_type_dict[name]

    def all_config_types(self):
        return self._config_type_dict.values()

    def has_context(self, name):
        check.str_param(name, 'name')
        return name in self.context_definitions

    def all_runtime_types(self):
        return self._runtime_type_dict.values()

    @property
    def solid_defs(self):
        return list(set([solid.definition for solid in self.solids]))

    def solid_def_named(self, name):
        check.str_param(name, 'name')

        for solid in self.solids:
            if solid.definition.name == name:
                return solid.definition

        check.failed('{} not found'.format(name))

    def has_solid_def(self, name):
        check.str_param(name, 'name')

        for solid in self.solids:
            if solid.definition.name == name:
                return True

        return False


def _create_adjacency_lists(solids, dep_structure):
    check.list_param(solids, 'solids', Solid)
    check.inst_param(dep_structure, 'dep_structure', DependencyStructure)

    visit_dict = {s.name: False for s in solids}
    forward_edges = {s.name: set() for s in solids}
    backward_edges = {s.name: set() for s in solids}

    def visit(solid_name):
        if visit_dict[solid_name]:
            return

        visit_dict[solid_name] = True

        for output_handle in dep_structure.deps_of_solid(solid_name):
            forward_node = output_handle.solid.name
            backward_node = solid_name
            if forward_node in forward_edges:
                forward_edges[forward_node].add(backward_node)
                backward_edges[backward_node].add(forward_node)
                visit(forward_node)

    for s in solids:
        visit(s.name)

    return (forward_edges, backward_edges)


def solids_in_topological_order(pipeline):
    check.inst_param(pipeline, 'pipeline', PipelineDefinition)

    _forward_edges, backward_edges = _create_adjacency_lists(
        pipeline.solids, pipeline.dependency_structure
    )

    order = toposort_flatten(backward_edges, sort=True)
    return [pipeline.solid_named(solid_name) for solid_name in order]



# # pylint: disable=W0622,W0614,W0401

from collections import namedtuple
import re

import pytest

from dagster import (
    DagsterRuntimeCoercionError,
    DagsterTypeError,
    DependencyDefinition,
    InputDefinition,
    OutputDefinition,
    PipelineDefinition,
    execute_pipeline,
    lambda_solid,
    make_dagster_type,
    solid,
)

from dagster.core.types.runtime import RuntimeType

StringTuple = make_dagster_type(namedtuple('StringTuple', 'str_one str_two'))


class SSNString(str):
    pass


class SSNStringTypeClass(RuntimeType):
    def __init__(self):
        super(SSNStringTypeClass, self).__init__(name='SSNString')

    def coerce_runtime_value(self, value):
        if isinstance(value, SSNString):
            return value

        if not isinstance(value, str):
            raise DagsterRuntimeCoercionError(
                '{value} is not a string. SSNStringType typecheck failed'.format(
                    value=repr(value)
                )
            )

        if not re.match(r'^(\d\d\d)-(\d\d)-(\d\d\d\d)$', value):
            raise DagsterRuntimeCoercionError(
                '{value} did not match SSN regex'.format(value=repr(value))
            )

        return SSNString(value)


@lambda_solid(output=OutputDefinition(StringTuple))
def produce_valid_value():
    return StringTuple(str_one='value_one', str_two='value_two')


@lambda_solid
def produce_invalid_value():
    return 'not_a_tuple'


@solid(inputs=[InputDefinition('string_tuple', StringTuple)])
def consume_string_tuple(info, string_tuple):
    info.context.info(
        'Logging value {string_tuple}'.format(string_tuple=string_tuple)
    )


@lambda_solid
def produce_valid_ssn_string():
    return '394-30-2032'


@lambda_solid
def produce_invalid_ssn_string():
    return '394-30-203239483'


@solid(inputs=[InputDefinition('ssn', SSNStringTypeClass)])
def consume_ssn(info, ssn):
    if not isinstance(ssn, SSNString):
        raise Exception('This should never be thrown')
    info.context.info('ssn: {ssn}'.format(ssn=ssn))


def define_part_twelve_step_one_pipeline():
    return PipelineDefinition(
        name='part_twelve_step_one_pipeline',
        solids=[produce_valid_value, consume_string_tuple],
        dependencies={
            'consume_string_tuple': {
                'string_tuple': DependencyDefinition('produce_valid_value')
            }
        },
    )


def define_part_twelve_step_two_pipeline():
    return PipelineDefinition(
        name='part_twelve_step_two_pipeline',
        solids=[produce_invalid_value, consume_string_tuple],
        dependencies={
            'consume_string_tuple': {
                'string_tuple': DependencyDefinition('produce_invalid_value')
            }
        },
    )


def define_part_twelve_step_three_pipeline():
    return PipelineDefinition(
        name='part_twelve_step_three_pipeline',
        solids=[produce_valid_ssn_string, consume_ssn],
        dependencies={
            'consume_ssn': {
                'ssn': DependencyDefinition('produce_valid_ssn_string')
            }
        },
    )


def define_part_twelve_step_four_pipeline():
    return PipelineDefinition(
        name='part_twelve_step_four_pipeline',
        solids=[produce_invalid_ssn_string, consume_ssn],
        dependencies={
            'consume_ssn': {
                'ssn': DependencyDefinition('produce_invalid_ssn_string')
            }
        },
    )


def test_ssn_type():
    good_ssn_string = '123-43-4939'
    good_ssn = SSNString(good_ssn_string)
    ssn_string_type = SSNStringTypeClass.inst()
    assert ssn_string_type.coerce_runtime_value(good_ssn_string) == good_ssn
    assert ssn_string_type.coerce_runtime_value(good_ssn) == good_ssn

    with pytest.raises(DagsterRuntimeCoercionError):
        ssn_string_type.coerce_runtime_value(123)

    with pytest.raises(DagsterRuntimeCoercionError):
        ssn_string_type.coerce_runtime_value(None)

    with pytest.raises(DagsterRuntimeCoercionError):
        ssn_string_type.coerce_runtime_value('12932-9234892038-384')

    with pytest.raises(DagsterRuntimeCoercionError):
        ssn_string_type.coerce_runtime_value('1292-34-383434')


def test_intro_tutorial_part_twelve_step_one():
    execute_pipeline(define_part_twelve_step_one_pipeline())


def test_intro_tutorial_part_twelve_step_two_pipeline():
    with pytest.raises(DagsterTypeError):
        execute_pipeline(define_part_twelve_step_two_pipeline())


def test_intro_tutorial_part_twelve_step_three_pipeline():
    execute_pipeline(define_part_twelve_step_three_pipeline())


def test_intro_tutorial_part_twelve_step_four():
    with pytest.raises(
        DagsterTypeError,
        match='Solid consume_ssn input ssn received value 394-30-203239483 ',
    ):
        execute_pipeline(define_part_twelve_step_four_pipeline())


if __name__ == '__main__':
    execute_pipeline(
        define_part_twelve_step_three_pipeline(), throw_on_error=True
    )



from collections import namedtuple

from dagster import check

from dagster.core.definitions import (
    solids_in_topological_order,
    InputDefinition,
    OutputDefinition,
    Solid,
    SolidOutputHandle,
)

from dagster.core.errors import DagsterInvariantViolationError

from .expectations import create_expectations_subplan, decorate_with_expectations

from .input_thunk import create_input_thunk_execution_step

from .materialization_thunk import decorate_with_output_materializations

from .objects import (
    ExecutionPlan,
    ExecutionPlanInfo,
    ExecutionStep,
    ExecutionValueSubPlan,
    ExecutionPlanSubsetInfo,
    StepInput,
    StepOutputHandle,
    StepTag,
)

from .transform import create_transform_step

from .utility import create_value_thunk_step


def get_solid_user_config(execution_info, pipeline_solid):
    check.inst_param(execution_info, 'execution_info', ExecutionPlanInfo)
    check.inst_param(pipeline_solid, 'pipeline_solid', Solid)

    name = pipeline_solid.name
    solid_configs = execution_info.environment.solids
    return solid_configs[name].config if name in solid_configs else None


# This is the state that is built up during the execution plan build process.
# steps is just a list of the steps that have been created
# step_output_map maps logical solid outputs (solid_name, output_name) to particular
# step outputs. This covers the case where a solid maps to multiple steps
# and one wants to be able to attach to the logical output of a solid during execution
StepBuilderState = namedtuple('StepBuilderState', 'steps step_output_map')


class StepOutputMap(dict):
    def __getitem__(self, key):
        check.inst_param(key, 'key', SolidOutputHandle)
        return dict.__getitem__(self, key)

    def __setitem__(self, key, val):
        check.inst_param(key, 'key', SolidOutputHandle)
        check.inst_param(val, 'val', StepOutputHandle)
        return dict.__setitem__(self, key, val)


def create_execution_plan_core(execution_info):
    check.inst_param(execution_info, 'execution_info', ExecutionPlanInfo)

    state = StepBuilderState(steps=[], step_output_map=StepOutputMap())

    for pipeline_solid in solids_in_topological_order(execution_info.pipeline):

        step_inputs = create_step_inputs(execution_info, state, pipeline_solid)

        solid_transform_step = create_transform_step(
            execution_info,
            pipeline_solid,
            step_inputs,
            get_solid_user_config(execution_info, pipeline_solid),
        )

        state.steps.append(solid_transform_step)

        for output_def in pipeline_solid.definition.output_defs:
            subplan = create_subplan_for_output(
                execution_info, pipeline_solid, solid_transform_step, output_def
            )
            state.steps.extend(subplan.steps)

            output_handle = pipeline_solid.output_handle(output_def.name)
            state.step_output_map[output_handle] = subplan.terminal_step_output_handle

    return create_execution_plan_from_steps(state.steps)


def create_execution_plan_from_steps(steps):
    check.list_param(steps, 'steps', of_type=ExecutionStep)

    step_dict = {step.key: step for step in steps}
    deps = {step.key: set() for step in steps}

    seen_keys = set()

    for step in steps:
        if step.key in seen_keys:
            keys = [s.key for s in steps]
            check.failed(
                'Duplicated key {key}. Full list: {key_list}.'.format(key=step.key, key_list=keys)
            )

        seen_keys.add(step.key)

        for step_input in step.step_inputs:
            deps[step.key].add(step_input.prev_output_handle.step.key)

    return ExecutionPlan(step_dict, deps)


def create_subplan_for_input(execution_info, solid, prev_step_output_handle, input_def):
    check.inst_param(execution_info, 'execution_info', ExecutionPlanInfo)
    check.inst_param(solid, 'solid', Solid)
    check.inst_param(prev_step_output_handle, 'prev_step_output_handle', StepOutputHandle)
    check.inst_param(input_def, 'input_def', InputDefinition)

    if execution_info.environment.expectations.evaluate and input_def.expectations:
        return create_expectations_subplan(
            solid, input_def, prev_step_output_handle, tag=StepTag.INPUT_EXPECTATION
        )
    else:
        return ExecutionValueSubPlan.empty(prev_step_output_handle)


def create_subplan_for_output(execution_info, solid, solid_transform_step, output_def):
    check.inst_param(execution_info, 'execution_info', ExecutionPlanInfo)
    check.inst_param(solid, 'solid', Solid)
    check.inst_param(solid_transform_step, 'solid_transform_step', ExecutionStep)
    check.inst_param(output_def, 'output_def', OutputDefinition)

    subplan = decorate_with_expectations(execution_info, solid, solid_transform_step, output_def)

    return decorate_with_output_materializations(execution_info, solid, output_def, subplan)


def get_input_source_step_handle(execution_info, state, solid, input_def):
    check.inst_param(execution_info, 'execution_info', ExecutionPlanInfo)
    check.inst_param(state, 'state', StepBuilderState)
    check.inst_param(solid, 'solid', Solid)
    check.inst_param(input_def, 'input_def', InputDefinition)

    input_handle = solid.input_handle(input_def.name)
    solid_config = execution_info.environment.solids.get(solid.name)
    dependency_structure = execution_info.pipeline.dependency_structure
    if solid_config and input_def.name in solid_config.inputs:
        input_thunk_output_handle = create_input_thunk_execution_step(
            execution_info, solid, input_def, solid_config.inputs[input_def.name]
        )
        state.steps.append(input_thunk_output_handle.step)
        return input_thunk_output_handle
    elif dependency_structure.has_dep(input_handle):
        solid_output_handle = dependency_structure.get_dep(input_handle)
        return state.step_output_map[solid_output_handle]
    else:
        raise DagsterInvariantViolationError(
            (
                'In pipeline {pipeline_name} solid {solid_name}, input {input_name} '
                'must get a value either (a) from a dependency or (b) from the '
                'inputs section of its configuration.'
            ).format(
                pipeline_name=execution_info.pipeline.name,
                solid_name=solid.name,
                input_name=input_def.name,
            )
        )


def create_step_inputs(info, state, solid):
    check.inst_param(info, 'info', ExecutionPlanInfo)
    check.inst_param(state, 'state', StepBuilderState)
    check.inst_param(solid, 'solid', Solid)

    step_inputs = []

    for input_def in solid.definition.input_defs:
        prev_step_output_handle = get_input_source_step_handle(info, state, solid, input_def)

        subplan = create_subplan_for_input(info, solid, prev_step_output_handle, input_def)

        state.steps.extend(subplan.steps)
        step_inputs.append(
            StepInput(input_def.name, input_def.runtime_type, subplan.terminal_step_output_handle)
        )

    return step_inputs


def create_subplan(execution_plan_info, execution_plan, subset_info):
    check.inst_param(execution_plan_info, 'execution_plan_info', ExecutionPlanInfo)
    check.inst_param(execution_plan, 'execution_plan', ExecutionPlan)
    check.inst_param(subset_info, 'subset_info', ExecutionPlanSubsetInfo)

    steps = []

    for step in execution_plan.steps:
        if step.key not in subset_info.subset:
            # Not included in subset. Skip.
            continue

        if step.key not in subset_info.inputs:
            steps.append(step)
        else:
            steps.extend(_create_new_steps_for_input(step, subset_info))

    return create_execution_plan_from_steps(steps)


def _create_new_steps_for_input(step, subset_info):
    new_steps = []
    new_step_inputs = []
    for step_input in step.step_inputs:
        if step_input.name in subset_info.inputs[step.key]:
            value_thunk_step_output_handle = create_value_thunk_step(
                step.solid,
                step_input.runtime_type,
                step.key + '.input.' + step_input.name + '.value',
                subset_info.inputs[step.key][step_input.name],
            )

            new_value_step = value_thunk_step_output_handle.step

            new_steps.append(new_value_step)

            new_step_inputs.append(
                StepInput(step_input.name, step_input.runtime_type, value_thunk_step_output_handle)
            )
        else:
            new_step_inputs.append(step_input)

    new_steps.append(step.with_new_inputs(new_step_inputs))
    return new_steps



# encoding: utf-8

from dagster import execute_pipeline
from dagster.tutorials.utils import (
    check_cli_execute_file_pipeline,
    check_script,
)
from dagster.utils import script_relative_path

from ..config import define_configurable_hello_pipeline


def test_tutorial_part_four():
    pipeline = define_configurable_hello_pipeline()

    result = execute_pipeline(
        pipeline, {'solids': {'configurable_hello': {'config': 'cn'}}}
    )

    assert result.success
    assert len(result.result_list) == 1
    assert (
        result.result_for_solid('configurable_hello').transformed_value()
        == '你好, 世界!'
    )
    return result


def test_tutorial_script_part_four():
    check_script(script_relative_path('../config.py'))


def test_tutorial_cli_part_four():
    check_cli_execute_file_pipeline(
        script_relative_path('../config.py'),
        'define_configurable_hello_pipeline',
        script_relative_path('../config_env.yml'),
    )
